import numpy as np
import numpy.typing as npt


def measure(emb: npt.NDArray, label: npt.NDArray) -> dict:
    """
    Compute distance consistency of the embedding
    INPUT:
        ndarray: emb: embedded data
        ndarray: label: label of the original data
    OUTPUT:
        dict: distance consistency (dsc)
    """

    # compute centroids
    point_num = emb.shape[0]
    label_num = np.unique(label).shape[0]

    centroids = np.zeros((label_num, emb.shape[1]))
    for i in range(label_num):
        centroids[i] = np.mean(emb[label == i], axis=0)

    # compute distance consistency
    consistent_num = 0
    for idx in range(point_num):
        current_label = -1
        current_dist = 1e10
        for c_idx in range(len(centroids)):
            dist = np.linalg.norm(emb[idx] - centroids[c_idx])
            if dist < current_dist:
                current_dist = dist
                current_label = c_idx
        if current_label == label[idx]:
            consistent_num += 1

    dsc = consistent_num / point_num
    return {"distance_consistency": dsc}



from .utils import pairwise_dist as pdist
from sklearn.isotonic import IsotonicRegression
import numpy as np
import numpy.typing as npt


def measure(
    orig: npt.NDArray, emb: npt.NDArray, distance_matrices: tuple | None = None
) -> dict:
    """
    Compute the non-metric stress of the embedding, described in:
    Kruskal, J. B. (1964). Multidimensional scaling by optimizing goodness of fit to a nonmetric hypothesis. Psychometrika, 29(1), 1-27.

    Or for a more modern description:
    Smelser, K., Miller, J., & Kobourov, S. (2024). " Normalized Stress" is Not Normalized: How to Interpret Stress Correctly. arXiv preprint arXiv:2408.07724.

    INPUT:
        ndarray: orig: orignal data
        ndarray: emb: embedded data
        tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
    OUTPUT:
        dict: non_metric_stress
    """
    if distance_matrices is None:
        orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        emb_distance_matrix = pdist.pairwise_distance_matrix(emb)

    else:
        orig_distance_matrix, emb_distance_matrix = distance_matrices

    # Extract upper triangular of both matrices into 1d arrays
    # Diagonal is always zero, we can offset by one
    N = orig_distance_matrix.shape[0]
    orig_upper_tri = orig_distance_matrix[np.triu_indices(N, k=1)]
    emb_upper_tri = emb_distance_matrix[np.triu_indices(N, k=1)]

    # Find the indices of orignal distance matrix that, when reordered, would sort it
    sorted_indices = np.argsort(orig_upper_tri)

    # Apply order to both arrays
    orig_upper_tri = orig_upper_tri[sorted_indices]
    emb_upper_tri = emb_upper_tri[sorted_indices]

    # Create the curve which minimizes horizontal distance in the Shepard diagram
    d_hat = (
        IsotonicRegression().fit(orig_upper_tri, emb_upper_tri).predict(orig_upper_tri)
    )

    # Finally, compute the stress of the horizontal distance between the fitted line and embedded distances
    raw_stress = np.sum(np.square(emb_upper_tri - d_hat))
    normalization_factor = np.sum(np.square(emb_upper_tri))

    non_metric_stress = np.sqrt(raw_stress / normalization_factor)

    return {"non_metric_stress": non_metric_stress}



import numpy as np
import numpy.typing as npt
from .utils import knn


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    k: int = 20,
    snn_info: tuple | None = None,
    knn_info: tuple | None = None,
) -> dict:
    """
    Compute neighbor dissimilarity (ND) of the embedding
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
            int: k: number of nearest neighbors to consider
            tuple: knn_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
    OUTPUT:
            dict: neighbor dissimilarity (ND)
    """

    if snn_info is None:
        if knn_info is None:
            orig_SNN_graph = knn.snn(orig, k, directed=False)
            emb_SNN_graph = knn.snn(emb, k, directed=False)
        else:
            orig_SNN_graph = knn.snn(orig, k, knn_indices=knn_info[0], directed=False)
            emb_SNN_graph = knn.snn(emb, k, knn_indices=knn_info[1], directed=False)

    else:
        orig_SNN_graph, emb_SNN_graph = snn_info

    D = (orig_SNN_graph - emb_SNN_graph) / k

    D_plus = D[D > 0]
    D_minus = D[D < 0]

    dissim_plus = np.sqrt(np.sum(D_plus**2))
    dissim_minus = np.sqrt(np.sum(D_minus**2))

    nd = max(dissim_plus, dissim_minus)

    return {"neighbor_dissimilarity": nd}



__all__ = [
    "class_aware_trustworthiness_continuity",
    "spearman_rho",
    "neighbor_dissimilarity",
    "distance_to_measure",
    "local_continuity_meta_criteria",
    "internal_validation_measure",
    "pearson_r",
    "distance_consistency",
    "kl_divergence",
    "neighborhood_hit",
    "trustworthiness_continuity",
    "clustering_and_external_validation_measure",
    "mean_relative_rank_error",
    "steadiness_cohesiveness",
    "topographic_product",
    "procrustes",
    "stress",
    "label_trustworthiness_and_continuity",
    "scale_normalized_stress",
    "non_metric_stress",
]

# add to __all__ if the function is added



from sklearn.cluster import KMeans, DBSCAN
from sklearn.metrics import (
    adjusted_rand_score,
    adjusted_mutual_info_score,
    normalized_mutual_info_score,
    v_measure_score,
)
import numpy.typing as npt


def measure(
    emb: npt.NDArray,
    label: npt.NDArray,
    measure: str = "arand",
    clustering: str = "kmeans",
    clustering_args=None,
) -> dict:
    """
    Evaluate DR embedding using clustering and external validation measure
    INPUT:
            ndarray: emb: embedded data
            str: clustering: clustering algorithm to use (Optional)
                    Currently supports "K-Means", "DBSCAN"
            dict: clustering_args: arguments for clustering algorithm (Optional)
            str: measure: external validation measure to compute (Optional)
                    Currently supports "arand", "ami", "nmi", "vmeasure"
    OUTPUT:
            dict: clustering and external validation measure value
    """
    measure = measure.lower()
    clustering = clustering.lower()

    if clustering_args is None:
        clustering_args = {}

    if clustering == "kmeans":
        clustering_result = KMeans(**clustering_args).fit(emb)
    elif clustering == "dbscan":
        clustering_result = DBSCAN(**clustering_args).fit(emb)
    else:
        raise ValueError("Invalid clustering algorithm")

    if measure == "arand":
        score = adjusted_rand_score(label, clustering_result.labels_)
    elif measure == "ami":
        score = adjusted_mutual_info_score(label, clustering_result.labels_)
    elif measure == "nmi":
        score = normalized_mutual_info_score(label, clustering_result.labels_)
    elif measure == "vmeasure":
        score = v_measure_score(label, clustering_result.labels_)

    return {f"{clustering}_{measure}": score}



import numpy as np
from .utils import knn
import numpy.typing as npt


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    label: npt.NDArray,
    k: int = 20,
    knn_ranking_info: tuple | None = None,
    return_local: bool = False,
) -> tuple | dict:
    """
    Compute class-aware trustworthiness and continuity of the embedding
    INPUT:
        ndarray: orig: original data
        ndarray: emb: embedded data
        ndarray: label: label of the original data
        int: k: number of nearest neighbors to consider
        tuple: knn_ranking_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
    OUTPUT:
        dict: class-aware trustworthiness (ca_trustworthiness) and class-aware continuity (ca_continuity)
    """

    if knn_ranking_info is None:
        orig_knn_indices, orig_ranking = knn.knn_with_ranking(orig, k)
        emb_knn_indices, emb_ranking = knn.knn_with_ranking(emb, k)
    else:
        orig_knn_indices, orig_ranking, emb_knn_indices, emb_ranking = knn_ranking_info

    if return_local:
        ca_trust, local_ca_trust = ca_tnc_computation(
            orig_knn_indices,
            orig_ranking,
            emb_knn_indices,
            label,
            k,
            "false",
            return_local,
        )
        ca_cont, local_ca_cont = ca_tnc_computation(
            emb_knn_indices,
            emb_ranking,
            orig_knn_indices,
            label,
            k,
            "missing",
            return_local,
        )
        return (
            {"ca_trustworthiness": ca_trust, "ca_continuity": ca_cont},
            {
                "local_ca_trustworthiness": local_ca_trust,
                "local_ca_continuity": local_ca_cont,
            },
        )
    else:
        ca_trust = ca_tnc_computation(
            orig_knn_indices,
            orig_ranking,
            emb_knn_indices,
            label,
            k,
            "false",
            return_local,
        )
        ca_cont = ca_tnc_computation(
            emb_knn_indices,
            emb_ranking,
            orig_knn_indices,
            label,
            k,
            "missing",
            return_local,
        )

        return {"ca_trustworthiness": ca_trust, "ca_continuity": ca_cont}


def ca_tnc_computation(
    base_knn_indices: npt.NDArray,
    base_ranking: npt.NDArray,
    target_knn_indices: npt.NDArray,
    label: npt.NDArray,
    k: int,
    type_description: str,
    return_local: bool = False,
) -> npt.NDArray | tuple:
    """
    Core computation of class-aware trustworthiness and continuity
    """

    local_distortion_list = []
    points_num = base_knn_indices.shape[0]

    for i in range(points_num):
        missings = np.setdiff1d(target_knn_indices[i], base_knn_indices[i])
        local_distortion = 0.0
        for missing in missings:
            if type_description == "false":
                if label[i] != label[missing]:
                    local_distortion += base_ranking[i, missing] - k
            elif type_description == "missing":
                if label[i] == label[missing]:
                    local_distortion += base_ranking[i, missing] - k
            else:
                raise ValueError("type should be 'false' or 'missing'")

        local_distortion_list.append(local_distortion)

    local_distortion_list = np.array(local_distortion_list)
    local_distortion_list = 1 - local_distortion_list * (
        2 / (k * (2 * points_num - 3 * k - 1))
    )

    average_distortion = np.mean(local_distortion_list)

    if return_local:
        return average_distortion, local_distortion_list
    else:
        return average_distortion



from scipy.spatial.distance import cdist
import numpy as np
import numpy.typing as npt
from typing import Callable


def pairwise_distance_matrix(
    point: npt.NDArray, distance_function: str | Callable = "euclidean"
):
    """
    Compute the pairwise distance matrix of the point list.
    You can use any distance function from scipy.spatial.distance.cdist
    or specify a callable function.

    INPUT:
        ndarray: point: list of points
        str or callable: distance_function: distance function to use
    OUTPUT:
        ndarray: pairwise distance matrix
    """
    if callable(distance_function):
        distance_matrix = cdist(point, point, distance_function)
    elif distance_function.lower() == "snn":
        # TODO
        raise NotImplementedError(
            "snn has not yet been implemented as a distance function"
        )
    else:
        distance_matrix = cdist(point, point, distance_function)

    return distance_matrix


def distance_matrix_to_density(
    distance_matrix: npt.NDArray, sigma: float
) -> npt.NDArray:
    """
    Compute the density of each point based on the pairwise distance matrix.

    INPUT:
        ndarray: distance_matrix: pairwise distance matrix
        float: sigma: sigma parameter for the Gaussian kernel
    OUTPUT:
        ndarray: density
    """
    normalized_distance_matrix = distance_matrix / np.max(distance_matrix)
    density = np.sum(np.exp(-(normalized_distance_matrix**2) / sigma), axis=-1)
    density = density / np.sum(density)
    return density



import numpy as np
import numpy.typing as npt
from .utils import knn


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    k: int = 20,
    knn_ranking_info: tuple | None = None,
    return_local: bool = False,
) -> tuple | dict:
    """
    Compute the trustworthiness and continuity of the embedding
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
            int: k: number of nearest neighbors to consider
            tuple: knn_ranking_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
    OUTPUT:
            dict: trustworthiness and continuity
    """
    if knn_ranking_info is None:
        orig_knn_indices, orig_ranking = knn.knn_with_ranking(orig, k)
        emb_knn_indices, emb_ranking = knn.knn_with_ranking(emb, k)
    else:
        orig_knn_indices, orig_ranking, emb_knn_indices, emb_ranking = knn_ranking_info

    if return_local:
        trust, local_trust = tnc_computation(
            orig_knn_indices, orig_ranking, emb_knn_indices, k, return_local
        )
        cont, local_cont = tnc_computation(
            emb_knn_indices, emb_ranking, orig_knn_indices, k, return_local
        )
        return (
            {"trustworthiness": trust, "continuity": cont},
            {"local_trustworthiness": local_trust, "local_continuity": local_cont},
        )
    else:
        trust = tnc_computation(
            orig_knn_indices, orig_ranking, emb_knn_indices, k, return_local
        )
        cont = tnc_computation(
            emb_knn_indices, emb_ranking, orig_knn_indices, k, return_local
        )
        return {"trustworthiness": trust, "continuity": cont}


def tnc_computation(
    base_knn_indices: npt.NDArray,
    base_ranking: npt.NDArray,
    target_knn_indices: npt.NDArray,
    k: int,
    return_local: bool = False,
) -> tuple | npt.NDArray:
    """
    Core computation of trustworthiness and continuity
    """
    local_distortion_list = []
    points_num = base_knn_indices.shape[0]

    for i in range(points_num):
        missings = np.setdiff1d(target_knn_indices[i], base_knn_indices[i])
        local_distortion = 0.0
        for missing in missings:
            local_distortion += base_ranking[i, missing] - k
        local_distortion_list.append(local_distortion)
    local_distortion_list = np.array(local_distortion_list)
    local_distortion_list = 1 - local_distortion_list * (
        2 / (k * (2 * points_num - 3 * k - 1))
    )

    average_distortion = np.mean(local_distortion_list)

    if return_local:
        return average_distortion, local_distortion_list
    else:
        return average_distortion



from .utils import pairwise_dist as pdist
import numpy as np
import numpy.typing as npt


def measure(
    orig: npt.NDArray, emb: npt.NDArray, distance_matrices: tuple | None = None
) -> dict:
    """
    Compute stress of the embedding
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
    tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
    OUTPUT:
            dict: stress
    """
    if distance_matrices is None:
        orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        emb_distance_matrix = pdist.pairwise_distance_matrix(emb)

    else:
        orig_distance_matrix, emb_distance_matrix = distance_matrices

    diff_squared_sum = np.square(orig_distance_matrix - emb_distance_matrix).sum()
    orig_squared_sum = np.square(orig_distance_matrix).sum()

    stress = np.sqrt(diff_squared_sum / orig_squared_sum)

    return {"stress": stress}



from .utils import pairwise_dist as pdist
import numpy as np
import numpy.typing as npt
from .utils import knn


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    k: int = 20,
    distance_matrices: tuple | None = None,
    knn_info: tuple | None = None,
) -> dict:
    """
    Compute topographic product
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
          int: k: number of nearest neighbors to consider
    OUTPUT:
            topographic product result
    """
    N = len(emb)
    sum_of_log_p3 = 0

    if distance_matrices is None:
        orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
    else:
        orig_distance_matrix, emb_distance_matrix = distance_matrices

    # k nearest neighbors in original space and embedded space each
    if knn_info is None:
        orig_knn_indices = knn.knn(orig, k)
        emb_knn_indices = knn.knn(emb, k)
    else:
        orig_knn_indices, emb_knn_indices = knn_info

    for j in range(N):
        for ki in range(k):
            q1_product, q2_product = 1, 1
            for l in range(ki):
                distance_origin_to_emb_knn = orig_distance_matrix[j][
                    emb_knn_indices[j][l]
                ]
                distance_origin_to_origin_knn = orig_distance_matrix[j][
                    orig_knn_indices[j][l]
                ]
                q1 = distance_origin_to_emb_knn / distance_origin_to_origin_knn
                q1_product *= q1

                distance_emb_to_emb_knn = emb_distance_matrix[j][emb_knn_indices[j][l]]
                distance_emb_to_origin_knn = emb_distance_matrix[j][
                    orig_knn_indices[j][l]
                ]
                q2 = distance_emb_to_emb_knn / distance_emb_to_origin_knn
                q2_product *= q2

            p3 = pow(q1_product * q2_product, 1 / (2 * (ki + 1)))
            sum_of_log_p3 += np.log(p3)

    topographic_product = sum_of_log_p3 / (N * k)
    return {"topographic_product": topographic_product}



from snc.snc import SNC
import numpy as np
import numpy.typing as npt

# from .utils import knn


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    iteration: int = 150,
    walk_num_ratio: float = 0.3,
    alpha: float = 0.1,
    k: int = 50,
    clustering_strategy: str = "dbscan",
    knn_info: tuple | None = None,
    return_local: bool = False,
) -> tuple | dict:
    """
    Compute the Steadiness and Cohesiveness of the embedding
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
            int: iteration: number of iterations for the SNC algorithm
            float: walk_num_ratio: ratio of the number of random walks to the number of points
            float: alpha: parameter for the SNC algorithm
            int: k: number of nearest neighbors to consider
            str: clustering_strategy: clustering strategy to use (dbscan or kmeans)
            tuple: knn_info: precomputed k-nearest neighbors of the original and embedded data (Optional)
    OUTPUT:
            dict: steadiness and cohesiveness score
    """

    # if knn_info is None:
    #     orig_knn_indices = knn.knn(orig, k)
    #     emb_knn_indices = knn.knn(emb, k)
    # else:
    #     orig_knn_indices, emb_knn_indices = knn_info

    # orig_snn_graph = knn.snn(orig, k, knn_indices=orig_knn_indices, directed=True)
    # emb_snn_graph = knn.snn(emb, k, knn_indices=emb_knn_indices, directed=True)

    # snn_knn_matrix = {
    #     "raw_knn": orig_knn_indices,
    #     "raw_snn": orig_snn_graph,
    #     "emb_knn": emb_knn_indices,
    #     "emb_snn": emb_snn_graph,
    # }

    snc_obj = SNC(
        orig,
        emb,
        iteration=iteration,
        walk_num_ratio=walk_num_ratio,
        dist_strategy="snn",
        dist_parameter={"alpha": alpha},
        dist_function=None,
        cluster_strategy=clustering_strategy,
        # snn_knn_matrix=snn_knn_matrix
    )

    snc_obj.fit(record_vis_info=return_local)

    steadiness = snc_obj.steadiness()
    cohesiveness = snc_obj.cohesiveness()

    if return_local:
        _, _, _, points_info = snc_obj.vis_info()

        stead_local = np.array(
            [1 - point_info["false_val"] for point_info in points_info]
        )
        cohev_local = np.array(
            [1 - point_info["missing_val"] for point_info in points_info]
        )

    if return_local:
        return {"steadiness": steadiness, "cohesiveness": cohesiveness}, {  # TODO
            "local_steadiness": stead_local,
            "local_cohesiveness": cohev_local,
        }
    else:
        return {"steadiness": steadiness, "cohesiveness": cohesiveness}



from scipy.stats import spearmanr
from .utils import pairwise_dist as pdist
import numpy.typing as npt


def measure(
    orig: npt.NDArray, emb: npt.NDArray, distance_matrices: tuple | None = None
) -> dict:
    """
    Compute Spearman's rank correlation coefficient of the embedding
    INPUT:
                  ndarray: orig: original data
                  ndarray: emb: embedded data
                  tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
          OUTPUT:
                  dict: Spearman's rank correlation coefficient (rho)
    """

    if distance_matrices is None:
        orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
    else:
        orig_distance_matrix, emb_distance_matrix = distance_matrices

    rho, p = spearmanr(orig_distance_matrix.flatten(), emb_distance_matrix.flatten())
    return {
        "spearman_rho": rho,
    }



from .pairwise_dist import pairwise_distance_matrix
import numpy as np
import numpy.typing as npt
import faiss
from sklearn.neighbors import KDTree
from scipy.sparse import csr_matrix


def knn_with_ranking(
    points: npt.NDArray, k: int, distance_matrix: npt.NDArray | None = None
) -> (npt.NDArray, npt.NDArray):
    """
    Compute the k-nearest neighbors of the points along with the
    rankings of other points based on the distance to each point.
    If the distance matrix is not provided, it is computed in O(n^2) time.
    INPUT:
        ndarray: points: list of points
        int: k: number of nearest neighbors to compute
        ndarray: distance_matrix: pairwise distance matrix (Optional)
    OUTPUT:
        ndarray: knn_indices: k-nearest neighbors of each point
        ndarray: ranking: ranking of other points based on the distance to each point
    """

    if distance_matrix is None:
        distance_matrix = pairwise_distance_matrix(points, "euclidean")

    sorted_indices = np.argsort(distance_matrix, axis=1)
    knn_indices = sorted_indices[:, 1 : k + 1]
    ranking = np.argsort(sorted_indices, axis=1)

    return knn_indices, ranking


def knn(
    points: npt.NDArray, k: int, distance_function: str = "euclidean"
) -> npt.NDArray:
    """
    Compute the k-nearest neighbors of the points
    If the distance function is euclidean, the computation relies on faiss-cpu.
    Otherwise, the computation is done based on scikit-learn KD Tree algorithm
    You can use any distance function supported by scikit-learn KD Tree or specify a callable function
    INPUT:
        ndarray: points: list of points
        int: k: number of nearest neighbors to compute
        str or callable: distance_function: distance function to use
    OUTPUT:
        ndarray: knn_indices: k-nearest neighbors of each point
    """

    # make c-contiguous
    points = np.ascontiguousarray(points, dtype=np.float32)

    if distance_function.lower() == "euclidean":
        index = faiss.IndexFlatL2(points.shape[1])
        index.add(points)
        knn_indices = index.search(points, k + 1)[1][:, 1:]
    else:
        tree = KDTree(points, metric=distance_function)
        knn_indices = tree.query(points, k=k + 1, return_distance=False)[:, 1:]

    return knn_indices


def snn(
    points: npt.NDArray,
    k: int,
    distance_function: str = "euclidean",
    directed: bool = True,
    knn_indices: tuple | None = None,
) -> npt.NDArray:
    """
    Compute the shared nearest neighbors (SNN) graph of the points
    INPUT:
        ndarray: points: list of points
        int: k: number of nearest neighbors to consider
        str or callable: distance_function: distance function to use
        bool: directed: whether the k-nearest neighbors graph using is directed or not
        tuple: knn_indices: precomputed k-nearest neighbors and rankings of the points (Optional)
    OUTPUT:
        ndarray: snn_graph: shared nearest neighbors (SNN) graph of the points
    """
    if knn_indices is None:
        knn_indices = knn(points, k, distance_function)

    n = knn_indices.shape[0]
    rows = np.repeat(np.arange(n), k)
    cols = knn_indices.flatten()
    vals = np.tile(np.arange(k, 0, -1), n)

    knn_graph = csr_matrix((vals, (rows, cols)), shape=(n, n))

    if directed:
        snn_graph = knn_graph @ knn_graph.T
    else:
        sym_graph = ((knn_graph + knn_graph.T) > 0).astype(int)
        snn_graph = sym_graph @ sym_graph

    snn_graph.setdiag(0)
    return snn_graph



from .measures import *
from .measures.utils import knn
from .measures.utils import pairwise_dist as pdist
import math
import numpy as np


class ZADU:
    ABBREVIATIONS = {
        "tnc": "trustworthiness_continuity",
        "mrre": "mean_relative_rank_error",
        "lcmc": "local_continuity_meta_criteria",
        "nh": "neighborhood_hit",
        "ca_tnc": "class_aware_trustworthiness_continuity",
        "l_tnc": "label_trustworthiness_and_continuity",
        "nd": "neighbor_dissimilarity",
        "dtm": "distance_to_measure",
        "kl_div": "kl_divergence",
        "dsc": "distance_consistency",
        "pr": "pearson_r",
        "srho": "spearman_rho",
        "ivm": "internal_validation_measure",
        "c_evm": "clustering_and_external_validation_measure",
        "snc": "steadiness_cohesiveness",
        "topo": "topographic_product",
        "proc": "procrustes",
        "stress": "stress",
    }

    DEFAULT_K = 20

    def __init__(
        self, spec_list, orig, return_local=False, verbose=False, geodesic=False
    ):
        self.spec_list = spec_list
        self.return_local = return_local
        self.verbose = verbose

        self.orig = orig
        self.emb = None
        self.label = None

        ## FLAGS for scheduling
        self.knn_flag = False
        self.knn_flag_k = -1
        self.knn_ranking_flag = False
        self.knn_ranking_flag_k = -1
        self.distance_matrices_flag = False

        ## variables for holding precomputed results (prerequisite for some measures)
        self.orig_distance_matrix = None
        self.emb_distance_matrix = None
        self.orig_knn_ranking = None
        self.emb_knn_ranking = None
        self.orig_knn_indices = None
        self.emb_knn_indices = None

        self.__sanity_check_measures_spec()
        self.__interpret_measures_spec()

        if self.distance_matrices_flag:
            if geodesic:
                self.orig_distance_matrix = self.__pairwise_geodesic_distance_matrix(
                    orig
                )
            else:
                self.orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        if self.knn_ranking_flag:
            self.orig_knn_indices, self.orig_knn_ranking = knn.knn_with_ranking(
                orig, self.knn_ranking_flag_k, distance_matrix=self.orig_distance_matrix
            )
        elif self.knn_flag and self.knn_flag_k > self.knn_ranking_flag_k:
            self.orig_knn_indices = knn.knn(
                orig, self.knn_flag_k, distance_function="euclidean"
            )

    def measure(self, emb, label=None):
        """
        Run the functions specified in spec_list
        INPUT:
                emb:  embedded data
        OUTPUT:
                list: list of results
        """

        self.emb = emb
        self.label = label

        ## compute the distance matrices
        if self.distance_matrices_flag:
            self.emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
        if self.knn_ranking_flag:
            self.emb_knn_indices, self.emb_knn_ranking = knn.knn_with_ranking(
                emb, self.knn_ranking_flag_k, distance_matrix=self.emb_distance_matrix
            )
        elif self.knn_flag and self.knn_flag_k > self.knn_ranking_flag_k:
            self.emb_knn_indices = knn.knn(
                emb, self.knn_flag_k, distance_function="euclidean"
            )

        ## compute the measures
        score_results = []
        local_results = []
        for spec in self.spec_list:
            measure_name = spec["id"]
            given_params = spec["params"] if "params" in spec else {}
            real_params = self.__get_real_params(measure_name)

            ## construct the execution parameters to be injected in the function
            exec_params = {}
            for param in given_params.keys():
                exec_params[param] = given_params[param]

            for param in real_params:
                if "orig" == param:
                    exec_params["orig"] = self.orig
                elif "emb" == param:
                    exec_params["emb"] = self.emb
                elif "label" == param:
                    if label is None:
                        raise Exception(
                            "Label is required for measure {}".format(measure_name)
                        )
                    exec_params["label"] = label
                elif "distance_matrices" == param:
                    exec_params["distance_matrices"] = (
                        self.orig_distance_matrix,
                        self.emb_distance_matrix,
                    )
                elif "knn_ranking_info" == param:
                    k_val = exec_params["k"] if "k" in exec_params else self.DEFAULT_K
                    exec_params["knn_ranking_info"] = (
                        self.orig_knn_indices[:, :k_val],
                        self.orig_knn_ranking,
                        self.emb_knn_indices[:, :k_val],
                        self.emb_knn_ranking,
                    )
                elif "knn_indices" == param:
                    k_val = exec_params["k"] if "k" in exec_params else self.DEFAULT_K
                    exec_params["knn_indices"] = (
                        self.orig_knn_indices[:, :k_val],
                        self.emb_knn_indices[:, :k_val],
                    )
                elif "return_local" == param:
                    exec_params["return_local"] = self.return_local

            ## execute the function
            if self.return_local and "return_local" in exec_params:
                score, local = globals()[measure_name].measure(**exec_params)
                score_results.append(score)
                local_results.append(local)
            elif self.return_local and "return_local" not in exec_params:
                score = globals()[measure_name].measure(**exec_params)
                score_results.append(score)
                local_results.append(None)
            else:
                score = globals()[measure_name].measure(**exec_params)
                score_results.append(score)

        if self.return_local:
            return score_results, local_results
        else:
            return score_results

    def __sanity_check_measures_spec(self):
        """
        Perform sanity check on the measures specification list.
        """
        ## check whehter there exists invalid measure name
        for spec in self.spec_list:
            if spec["id"] not in self.ABBREVIATIONS.values():
                if spec["id"] in self.ABBREVIATIONS:
                    spec["id"] = self.ABBREVIATIONS[spec["id"]]
                else:
                    raise Exception("Invalid measure name: {}".format(spec["id"]))

        ## check whether the parameters are valid
        for spec in self.spec_list:
            measure_name = spec["id"]
            given_params = spec["params"] if "params" in spec else {}
            real_params = self.__get_real_params(measure_name)

            ## check whether the given parameters are valid
            for param in given_params:
                if param not in real_params:
                    raise Exception(
                        f"Invalid parameter {param} for measure {measure_name}"
                    )

    def __interpret_measures_spec(self):
        """
        Interpret the measures spec and specify the preprequisites (knn, distance matrices)
        """
        for spec in self.spec_list:
            measure_name = spec["id"]
            given_params = spec["params"] if "params" in spec else {}
            real_params = self.__get_real_params(measure_name)

            if "knn_ranking_info" in real_params:
                self.knn_ranking_flag = True
                if "k" in given_params:
                    self.knn_ranking_flag_k = max(
                        self.knn_ranking_flag_k, given_params["k"]
                    )
                else:
                    self.knn_ranking_flag_k = max(
                        self.knn_ranking_flag_k, self.DEFAULT_K
                    )
            if "knn_info" in real_params:
                self.knn_flag = True
                if "k" in given_params:
                    self.knn_flag_k = max(self.knn_flag_k, given_params["k"])
                else:
                    self.knn_flag_k = max(self.knn_flag_k, self.DEFAULT_K)
            if "distance_matrices" in real_params:
                self.distance_matrices_flag = True

    def __get_real_params(self, measure_name):
        """
        Get the real parameters of a measure.
        """
        measure_func = globals()[measure_name].measure
        return measure_func.__code__.co_varnames[: measure_func.__code__.co_argcount]

    def __geodesic_distance(self, phi1, lambda1, phi2, lambda2):
        return math.acos(
            math.sin(phi1) * math.sin(phi2)
            + math.cos(phi1) * math.cos(phi2) * math.cos(abs(lambda2 - lambda1))
        )

    def __pairwise_geodesic_distance_matrix(self, orig):
        data_len = len(orig)
        distance_matrix = np.zeros((data_len, data_len))
        for i in range(data_len):
            for j in range(i, data_len):
                if i == j:
                    distance_matrix[i][j] = 0
                else:
                    distance_matrix[i][j] = distance_matrix[j][
                        i
                    ] = self.__geodesic_distance(
                        orig[i][1], orig[i][0], orig[j][1], orig[j][0]
                    )
        return distance_matrix



import numpy as np


def cielab_to_rgb_hex(L, a, b):
    # Convert CIELAB to XYZ
    def lab_to_xyz(L, a, b):
        y = (L + 16) / 116
        x = a / 500 + y
        z = y - b / 200

        x = x**3 if x**3 > 0.008856 else (x - 16 / 116) / 7.787
        y = y**3 if y**3 > 0.008856 else (y - 16 / 116) / 7.787
        z = z**3 if z**3 > 0.008856 else (z - 16 / 116) / 7.787

        x = x * 95.047
        y = y * 100
        z = z * 108.883

        return x, y, z

    # Convert XYZ to RGB
    def xyz_to_rgb(x, y, z):
        x /= 100
        y /= 100
        z /= 100

        r = x * 3.2406 + y * -1.5372 + z * -0.4986
        g = x * -0.9689 + y * 1.8758 + z * 0.0415
        b = x * 0.0557 + y * -0.2040 + z * 1.0570

        def convert(color):
            color = np.clip(color, 0, 1)
            color = np.where(
                color <= 0.0031308,
                12.92 * color,
                1.055 * np.power(color, 1 / 2.4) - 0.055,
            )
            return np.round(color * 255).astype(int)

        r, g, b = convert(r), convert(g), convert(b)
        return r, g, b

    x, y, z = lab_to_xyz(L, a, b)
    r, g, b = xyz_to_rgb(x, y, z)
    return f"#{r:02x}{g:02x}{b:02x}"


def checkviz_cmap(dist_false, dist_missing):
    ## generate a 2D continous colormap for the missing and false distortion
    """
    Generate a matplotlib color code that reproduces the above js code
    """
    cScale = 1.3
    dist_false = 1 - dist_false
    dist_missing = 1 - dist_missing

    powScale = lambda x: x**1.5145
    aScale = lambda x: 30 * cScale * x
    bScale = lambda x: 20 * cScale * x

    lab = [
        powScale(1 - (dist_false + dist_missing) / 2) * 100,
        aScale(dist_false - dist_missing),
        bScale(dist_missing - dist_false),
    ]
    ## change the cielab color to rgb that can be used by matplotlib
    # print(lab)
    color = cielab_to_rgb_hex(*lab)
    return color



import matplotlib.pyplot as plt
from scipy.spatial import Voronoi, voronoi_plot_2d
from .colormap import checkviz_cmap

from sklearn.neighbors import kneighbors_graph


def checkviz(
    scatter_data,
    false_distortion_list,
    missing_distortion_list,
    ax=None,
    point_c="black",
    point_s=1,
    point_alpha=0.5,
    point_marker="o",
):
    vor = Voronoi(scatter_data)
    ## set size
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 10))

    voronoi_plot_2d(vor, ax=ax, show_vertices=False, show_points=False, line_width=0)

    for idx, region in enumerate(vor.regions[:-1]):
        if not -1 in region:
            polygon = [vor.vertices[i] for i in region]
            ax.fill(
                *zip(*polygon),
                checkviz_cmap(false_distortion_list[idx], missing_distortion_list[idx])
            )

    ax.scatter(
        scatter_data[:, 0],
        scatter_data[:, 1],
        c=point_c,
        zorder=2,
        s=point_s,
        alpha=point_alpha,
        marker=point_marker,
    )

    ax.set_xticks([])
    ax.set_yticks([])


def reliability_map(
    emb,
    false_distortion_list,
    missing_distortion_list,
    k=7,
    ax=None,
    point_c="black",
    point_s=1,
    point_alpha=0.5,
    point_marker="o",
    linewidth=2.5,
    line_alpha=0.8,
):
    ## construct a knn graph
    if ax is None:
        fig, ax = plt.subplots(figsize=(10, 10))

    knn_graph = kneighbors_graph(emb, k, mode="distance", include_self=False)

    ## visualizae points and knn graph
    ax.scatter(
        emb[:, 0],
        emb[:, 1],
        c=point_c,
        zorder=2,
        s=point_s,
        alpha=point_alpha,
        marker=point_marker,
    )
    for i in range(emb.shape[0]):
        for j in knn_graph[i].indices:
            color = checkviz_cmap(
                (false_distortion_list[i] + false_distortion_list[j]) / 2,
                (missing_distortion_list[i] + missing_distortion_list[j]) / 2,
            )
            ax.plot(
                [emb[i, 0], emb[j, 0]],
                [emb[i, 1], emb[j, 1]],
                c=color,
                zorder=1,
                linewidth=linewidth,
                alpha=line_alpha,
            )

    ax.set_xticks([])
    ax.set_yticks([])



import numpy as np
import numpy.typing as npt
from .utils import label_cvms as lcvms


def measure(
    orig: npt.NDArray, emb: npt.NDArray, label: npt.NDArray, cvm: str = "dsc"
) -> dict:
    """
    Compute label-trustworthiness and label-continuity of the embedding
    INPUT:
            ndarray: emb: embedded data
            ndarray: label: label of the original data
            str: measure: clustering validation measure (CVM) to compute (Optional)
                    Currently supports "dsc" (Distance Consistency)
    OUTPUT:
            dict: label-trustworthiness and label-continuity
    """

    orig = np.array(orig)
    emb = np.array(emb)
    label = np.array(label)

    # change label into 0, 1, 2,....
    unique_labels = np.unique(label)
    label_dict = {}
    for i, label_single in enumerate(unique_labels):
        label_dict[label_single] = i

    int_labels = np.zeros(label.shape[0])
    for i in range(label.shape[0]):
        int_labels[i] = label_dict[label[i]]
    label_num = len(unique_labels)

    cvm = {"dsc": lcvms.dsc_normalize, "ch_btw": lcvms.btw_ch}[cvm]

    # compute the label-pairwise cvm of the original data
    raw_cvm_mat = np.zeros((label_num, label_num))
    emb_cvm_mat = np.zeros((label_num, label_num))

    for label_i in range(label_num):
        for label_j in range(label_i + 1, label_num):
            # raw data of a pair of labels
            filter_label = np.logical_or(int_labels == label_i, int_labels == label_j)
            raw_pair = orig[filter_label]
            emb_pair = emb[filter_label]
            # label of the raw data of a pair of labels
            raw_pair_label = int_labels[filter_label]
            emb_pair_label = int_labels[filter_label]

            # change the label to 0 and 1
            raw_pair_label[raw_pair_label == label_i] = 0
            raw_pair_label[raw_pair_label == label_j] = 1
            emb_pair_label[emb_pair_label == label_i] = 0
            emb_pair_label[emb_pair_label == label_j] = 1

            # compute cvm
            raw_cvm_mat[label_i, label_j] = cvm(raw_pair, raw_pair_label)
            emb_cvm_mat[label_i, label_j] = cvm(emb_pair, emb_pair_label)

    # compute the label-trustworthiness and label-continuity score
    lt_mat = raw_cvm_mat - emb_cvm_mat
    lt_mat[lt_mat < 0] = 0
    lt = 1 - np.sum(lt_mat) / (label_num * (label_num - 1) / 2)

    lc_mat = emb_cvm_mat - raw_cvm_mat
    lc_mat[lc_mat < 0] = 0
    lc = 1 - np.sum(lc_mat) / (label_num * (label_num - 1) / 2)

    # set the dictionary to return
    return {"label_trustworthiness": lt, "label_continuity": lc}



from sklearn.metrics import (
    silhouette_score,
    calinski_harabasz_score,
    davies_bouldin_score,
)
import numpy.typing as npt


def measure(emb: npt.NDArray, label: npt.NDArray, measure: str = "silhouette") -> dict:
    """
    Compute internal validation measure of the embedding
    INPUT:
        ndarray: emb: embedded data
        str: measure: internal validation measure to compute (Optional) Currently supports "silhouette", "calinski_harabasz", "davies_bouldin"
    OUTPUT:
        dict: internal validation measure value
    """

    measure = measure.lower()

    if measure == "silhouette":
        score = silhouette_score(emb, label)
    elif measure == "calinski_harabasz":
        score = calinski_harabasz_score(emb, label)
    elif measure == "davies_bouldin":
        score = davies_bouldin_score(emb, label)

    return {measure: score}



from .utils import pairwise_dist as pdist
import numpy as np
import numpy.typing as npt
from .stress import measure as stressmeasure


def measure(
    orig: npt.NDArray, emb: npt.NDArray, distance_matrices: tuple | None = None
) -> dict:
    """
    Compute the minimum stress for a given embedding.
    INPUT:
        ndarray: orig: original data
        ndarray: emb: embedded data
        tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
    OUTPUT:
        dict: non_metric_stress
    """
    if distance_matrices is None:
        orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        emb_distance_matrix = pdist.pairwise_distance_matrix(emb)

    else:
        orig_distance_matrix, emb_distance_matrix = distance_matrices

    alpha = np.sum(np.multiply(orig_distance_matrix, emb_distance_matrix)) / np.sum(
        np.square(emb_distance_matrix)
    )
    sns = stressmeasure(
        orig, alpha * emb, (orig_distance_matrix, alpha * emb_distance_matrix)
    )

    return {"scale_normalized_stress": sns["stress"]}



from .utils import knn
import numpy as np
import numpy.typing as npt


def measure(
    emb: npt.NDArray,
    label: npt.NDArray,
    k: int = 20,
    knn_emb_info: tuple | None = None,
    return_local: bool = False,
) -> tuple | dict:
    """
    Compute neighborhood hit of the embedding
    INPUT:
            ndarray: emb: embedded data
            ndarray: label: label of the original data
            int: k: number of nearest neighbors to consider
            tuple: knn_info: precomputed k-nearest neighbors of the original and embedded data (Optional)
    OUTPUT:
            dict: neighborhood hit (nh)
    """
    if knn_emb_info is None:
        emb_knn_indices = knn.knn(emb, k)
    else:
        emb_knn_indices = knn_emb_info

    points_num = emb.shape[0]
    nh_list = []
    for i in range(points_num):
        emb_knn_index = emb_knn_indices[i]
        emb_knn_index_label = label[emb_knn_index]
        nh_list.append(np.sum((emb_knn_index_label == label[i]).astype(int)))

    nh_list = np.array(nh_list)
    nh_list = nh_list / k

    nh = np.mean(nh_list)

    if return_local:
        return ({"neighborhood_hit": nh}, {"local_neighborhood_hit": nh_list})
    else:
        return {"neighborhood_hit": nh}



import numpy as np
import numpy.typing as npt
from .utils import knn


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    k: int = 20,
    knn_ranking_info: tuple | None = None,
    return_local: bool = False,
) -> tuple | dict:
    """
    Compute Mean Relative Rank Error (MRRE) of the embedding
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
            int: k: number of nearest neighbors to consider
            tuple: knn_ranking_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
    OUTPUT:
            dict: MRRE_false and MRRE_missing
    """
    if knn_ranking_info is None:
        orig_knn_indices, orig_ranking = knn.knn_with_ranking(orig, k)
        emb_knn_indices, emb_ranking = knn.knn_with_ranking(emb, k)
    else:
        orig_knn_indices, orig_ranking, emb_knn_indices, emb_ranking = knn_ranking_info

    if return_local:
        mrre_false, local_mrre_false = mrre_computation(
            orig_ranking, emb_ranking, emb_knn_indices, k, return_local
        )
        mrre_missing, local_mrre_missing = mrre_computation(
            emb_ranking, orig_ranking, orig_knn_indices, k, return_local
        )
        return (
            {"mrre_false": mrre_false, "mrre_missing": mrre_missing},
            {
                "local_mrre_false": local_mrre_false,
                "local_mrre_missing": local_mrre_missing,
            },
        )
    else:
        mrre_false = mrre_computation(
            orig_ranking, emb_ranking, emb_knn_indices, k, return_local
        )
        mrre_missing = mrre_computation(
            emb_ranking, orig_ranking, orig_knn_indices, k, return_local
        )

        return {
            "mrre_false": mrre_false,
            "mrre_missing": mrre_missing,
        }


def mrre_computation(
    base_ranking: npt.NDArray,
    target_ranking: npt.NDArray,
    target_knn_indices: npt.NDArray,
    k: int,
    return_local: bool = False,
) -> tuple | dict:
    """
    Core computation of MRRE
    """
    local_distortion_list = []
    points_num = target_knn_indices.shape[0]
    for i in range(points_num):
        base_rank_arr = base_ranking[i][target_knn_indices[i]]
        target_rank_arr = target_ranking[i][target_knn_indices[i]]
        local_distortion_list.append(
            np.sum(np.abs(base_rank_arr - target_rank_arr) / target_rank_arr)
        )

    c = sum([abs(points_num - 2 * i + 1) / i for i in range(1, k + 1)])
    local_distortion_list = np.array(local_distortion_list)
    local_distortion_list = 1 - local_distortion_list / c

    average_distortion = np.mean(local_distortion_list)

    if return_local:
        return average_distortion, local_distortion_list
    else:
        return average_distortion



import numpy as np
import numpy.typing as npt
from numpy.linalg import svd
from .utils import knn


def measure(
    orig: npt.NDArray, emb: npt.NDArray, k: int = 20, knn_info: tuple | None = None
) -> dict:
    """
    Compute procrustes statistics
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
          int: k: number of nearest neighbors to consider
    OUTPUT:
            Procrustes score
    """
    # k nearest neighbors in original space and embedded space each
    if knn_info is None:
        orig_knn_indices = knn.knn(orig, k)
        emb_knn_indices = knn.knn(emb, k)
    else:
        orig_knn_indices, emb_knn_indices = knn_info

    g_list = []

    for i in range(orig.shape[0]):
        origin_neighbors = orig[orig_knn_indices[i]]
        embedd_neighbors = emb[emb_knn_indices[i]]

        k = origin_neighbors.shape[0]
        I = np.eye(k)
        ones = np.ones((k, k))
        H = I - (1 / k) * ones

        U, _, V_T = svd(origin_neighbors.T @ H @ embedd_neighbors, full_matrices=False)
        A = U @ V_T
        A_T = A.T

        g = (
            np.linalg.norm(H @ (origin_neighbors - embedd_neighbors @ A_T), ord="fro")
            ** 2
        )
        g_normalized = g / np.linalg.norm(H @ origin_neighbors, ord="fro") ** 2
        g_list.append(g_normalized)

    procrustes = np.mean(g_list)

    return {"procrustes": procrustes}



from .utils import pairwise_dist as pdist
import numpy as np
import numpy.typing as npt


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    sigma: float = 0.1,
    distance_matrices: tuple | None = None,
) -> dict:
    """
    Compute Kullback-Leibler divergence of the embedding
    INPUT:
        ndarray: orig: original data
        ndarray: emb: embedded data
        tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
    OUTPUT:
        dict: Kullback-Leibler divergence (kl)
    """

    if distance_matrices is None:
        orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
    else:
        orig_distance_matrix, emb_distance_matrix = distance_matrices

    density_orig = pdist.distance_matrix_to_density(orig_distance_matrix, sigma)
    density_emb = pdist.distance_matrix_to_density(emb_distance_matrix, sigma)

    kl = np.sum(density_orig * np.log(density_orig / density_emb))
    return {"kl_divergence": kl}



from scipy.stats import pearsonr
from .utils import pairwise_dist as pdist
import numpy.typing as npt


def measure(
    orig: npt.NDArray, emb: npt.NDArray, distance_matrices: tuple | None = None
):
    """
    Compute Pearson's correlation coefficient of the embedding
    INPUT:
            ndarray: orig: original data
            ndarray: emb: embedded data
            tuple: distance_matrices: precomputed distance matrices of the original and embedded data (Optional)
    OUTPUT:
            dict: Pearson's correlation coefficient (r)
    """

    if distance_matrices is None:
        orig_distance_matrix = pdist.pairwise_distance_matrix(orig)
        emb_distance_matrix = pdist.pairwise_distance_matrix(emb)
    else:
        orig_distance_matrix, emb_distance_matrix = distance_matrices

    r, p = pearsonr(orig_distance_matrix.flatten(), emb_distance_matrix.flatten())
    return {
        "pearson_r": r,
    }



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

print(setuptools.find_packages(where="src"))

setuptools.setup(
    name="zadu",
    version="0.3.0",
    author="Hyeon Jeon",
    author_email="hj@hcil.snu.ac.kr",
    description="A Python Toolkit for Evaluating the Reliability of Dimensionality Reduction Embeddings",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hj-n/zadu",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "scikit-learn",
        "scipy",
        "snc",
        "matplotlib",
        "faiss-cpu",
    ],
    package_dir={"": "src"},
    packages=["zadu", "zaduvis", "zadu.measures", "zadu.measures.utils"],
    python_requires=">=3.10.0",
)



import numpy as np
import numpy.typing as npt
from .utils import knn


def measure(
    orig: npt.NDArray,
    emb: npt.NDArray,
    k: int = 20,
    knn_info: tuple | None = None,
    return_local: bool = False,
) -> tuple | dict:
    """
    Compute the local continuity meta-criteria of the embedding
    INPUT:
                  ndarray: orig: original data
                  ndarray: emb: embedded data
                  int: k: number of nearest neighbors to consider
                  tuple: knn_info: precomputed k-nearest neighbors and rankings of the original and embedded data (Optional)
    OUTPUT:
                  dict: local continuity meta-criteria
    """
    if knn_info is None:
        orig_knn_indices = knn.knn(orig, k)
        emb_knn_indices = knn.knn(emb, k)
    else:
        orig_knn_indices, emb_knn_indices = knn_info

    point_num = orig.shape[0]
    local_distortion_list = []

    for i in range(point_num):
        local_distortion_list.append(
            np.intersect1d(orig_knn_indices[i], emb_knn_indices[i]).shape[0]
            - ((k * k) / (point_num - 1))
        )

    local_distortion_list = np.array(local_distortion_list)
    local_distortion_list = local_distortion_list / k

    average_distortion = np.mean(local_distortion_list)

    if return_local:
        return ({"lcmc": average_distortion}, {"local_lcmc": local_distortion_list})
    else:
        return {"lcmc": average_distortion}



import numpy as np
from itertools import combinations
import numpy.typing as npt


def btw_ch(data: npt.NDArray, labels: npt.NDArray) -> float:
    return btw(data, labels)


def dsc_normalize(data: npt.NDArray, labels: npt.NDArray) -> float:
    """
    Compute the distance consistency in a vectorized manner.
    """
    labels = np.asarray(labels)
    unique_labels, labels = np.unique(labels, return_inverse=True)
    n_classes = len(unique_labels)

    # Compute centroids
    centroids = np.zeros((n_classes, data.shape[1]), dtype=float)
    for c in range(n_classes):
        centroids[c] = data[labels == c].mean(axis=0)

    dists = np.linalg.norm(data[:, None, :] - centroids[None, ...], axis=2)
    pred_labels = np.argmin(dists, axis=1)
    consistent_ratio = np.mean(pred_labels == labels)

    return (consistent_ratio - 0.5) * 2


def shift(X: npt.NDArray, label: npt.NDArray) -> float:
    labels = np.asarray(label)
    unique_labels, labels = np.unique(labels, return_inverse=True)
    n_clusters = len(unique_labels)
    n_samples, n_features = X.shape

    global_centroid = X.mean(axis=0)
    dists_to_global = np.linalg.norm(X - global_centroid, axis=1)
    std = dists_to_global.std()
    sums = np.vstack(
        [
            np.bincount(labels, weights=X[:, j], minlength=n_clusters)
            for j in range(n_features)
        ]
    ).T
    counts = np.bincount(labels, minlength=n_clusters)[:, None]
    centroids = sums / counts

    centroid_dists = np.linalg.norm(centroids - global_centroid, axis=1)

    dists_to_centroids = np.linalg.norm(X - centroids[labels], axis=1)
    compactness = np.sum(np.exp(dists_to_centroids / std))

    separability = np.sum(np.exp(centroid_dists / std) * counts.ravel())

    result = (separability * (n_samples - 2)) / compactness

    if not np.isfinite(result):
        raise FloatingPointError("Result became inf or NaN")

    return float(result)


def shift_range(X, label, iter_num):
    orig = shift(X, label)
    orig_result = 1 / (1 + orig ** (-1))
    e_val_sum = 0
    for i in range(iter_num):
        np.random.shuffle(label)
        e_val_sum += shift(X, label)
    e_val = e_val_sum / iter_num
    e_val_result = 1 / (1 + e_val ** (-1))
    if e_val_result == 1:
        return 0
    return (orig_result - e_val_result) / (1 - e_val_result)


def shift_range_class(X: npt.NDArray, label: npt.NDArray, iter_num: int) -> float:
    unique_labels = np.unique(label)
    class_indices = {c: np.where(label == c)[0] for c in unique_labels}

    def process_pair(a, b):
        idx = np.concatenate([class_indices[a], class_indices[b]])
        X_pair = X[idx]
        labels_pair = label[idx]
        labels_pair = (labels_pair == b).astype(np.int32)
        return shift_range(X_pair, labels_pair, iter_num)

    scores = [process_pair(a, b) for a, b in combinations(unique_labels, 2)]

    return float(np.mean(scores))


def btw(X: npt.NDArray, labels: npt.NDArray, iter_num: int = 20) -> float:
    return shift_range_class(X, labels, iter_num)


def centroid(X: npt.NDArray) -> npt.NDArray:
    """
    Compute the centroid of a set of vectors.
    :param X: The set of vectors.
    :return: The centroid.
    """
    return np.mean(X, axis=0)
